/**
 * Copyright (C) Voltaire Ltd. 2001-2010.  ALL RIGHTS RESERVED.
 * This software product is a proprietary product of Voltaire Ltd.
 * (the "Company") and all right, title, and interest and to the software product,
 * including all associated intellectual property rights, are and shall
 * remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 * $COPYRIGHT$
 * $HEADER$
 */


#include <errno.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <fca/fca_api.h>
#include <fca/config/fca_parse_specfile.h>
#include <assert.h>

#include "reduce_test.h"

#define USEC_PER_SEC	1000000L

#define info(fmt, arg...)	printf("REDUCE_TEST: "fmt"\n", ## arg)
#define error(fmt, arg...)	fprintf(stderr, "REDUCE_TEST: "fmt"\n", ## arg)

int g_progress_counter = 0;

/*
 * Progress function for the test
 * Will be called from FCA collective operation context periodically if FCA blocks
 * there for too long.
 */
void my_progress(void *arg)
{
	int *counter = arg;
	++(*counter);
}

/**
 * INIT1:
 * 
 * Initialize FCA context and select IB device
 */
struct fca_context *init_context(fca_init_spec_t  *fca_spec)
{
	fca_t *ctx;
	int ret;

	fca_spec->element_type = FCA_ELEMENT_RANK;
	fca_spec->progress.func = my_progress;
	fca_spec->progress.arg = &g_progress_counter;
	fca_spec->job_id = 0;
	ret = fca_init(fca_spec, &ctx);
	if (ret<0) {
		error("Error initializing FCA context, reason: %s", fca_strerror(ret));
		return NULL;
	}
	return ctx;
}

/**
 * Collect information from other ranks and get COMM_MLID from FMM
 */
int server_create_new_comm(struct fca_context *ctx, int port,
                          fca_comm_desc_t *comm_desc, int num_clients)
{
	int sockfd;
	int i, pos, ret = -1;
	char *all_info;
	void *my_info;
	int my_info_len, max_info_len;
	struct sockaddr_in local_addr = {0,};
	int connections[num_clients];
	int num_node_roots;
	fca_comm_new_spec_t spec;
	int info_len;
	int fin;
	int optval;

	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (sockfd < 0) {
		error("socket() failed: %m");
		return sockfd;
	}

	optval = 1;
	setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));

	/* Get my data */
	my_info = fca_get_rank_info(ctx, &my_info_len);
	if (!my_info) {
		error("fca_get_rank_info() returned NULL");
		goto out_close_srv;
	}

	/* Allocate total data */
	max_info_len = my_info_len * (num_clients + 1);
	all_info = malloc(max_info_len);
	if (!all_info) {
		error("malloc() failed: %m");
		goto out_free_my_info;
	}
	
	/* Copy my data to total data */
	memcpy(all_info, my_info, my_info_len);
	pos = my_info_len;
	
	/* Listen for connections */
	local_addr.sin_family = AF_INET;
	local_addr.sin_port = htons(port);
	local_addr.sin_addr.s_addr = INADDR_ANY;
	ret = bind(sockfd, (struct sockaddr *)&local_addr, sizeof local_addr);
	if (ret < 0) {
		error("bind() failed: %m");
		goto out_free_all_info;
	}
	
	ret = listen(sockfd, 100);
	if (ret < 0) {
		error("listen() failed: %m");
		goto out_free_all_info;
	}

	/* Reset numbers */
	for (i = 0; i < num_clients; ++i) {
		connections[i] = -1;
	}
	
	info("Waiting for %d connections..", num_clients);

	/* Collect information from other ranks */
	num_node_roots = 1;
	for (i = 0; i < num_clients; ++i) {
		ret = accept(sockfd, NULL, NULL);
		if (ret < 0) {
			error("accept() failed: %m");
			goto out_close_conns;
		}
		connections[i] = ret;

		/* Receive length */
		ret = recv(connections[i], &info_len, sizeof(info_len), 0);
		if (ret != sizeof(info_len)) {
			error("recv() failed: %m");
			goto out_close_conns;
		}

		assert(pos + info_len <= max_info_len);

		/* Receive data */
		if (info_len > 0) {
			ret = recv(connections[i], all_info + pos, info_len, 0);
			if (ret != info_len) {
				error("recv() failed: %m");
				goto out_close_conns;
			}

			++num_node_roots;
		}

		info("got info of %d bytes from rank  #%d (total: %d)", info_len, i, num_clients);
	}
	
	info("Creating new communicator with %d node roots...", num_node_roots);

	/* Create new communicator */
	spec.rank_info = all_info;
	spec.rank_count = num_node_roots;
	spec.is_comm_world = 1;

	ret = fca_comm_new(ctx, &spec, comm_desc);
	if (ret < 0) {
		error("fca_comm_new() failed: %s", fca_strerror(ret));
		goto out_close_conns;
	}

	info("Got communicator");

	/* Distribute information and close connections*/
	for (i = 0; i < num_clients; ++i) {
		ret = send(connections[i], comm_desc, sizeof *comm_desc, 0);
		if (ret < 0) {
			error("send() failed: %m");
			goto out_close_conns;
		}
	}

	/* receive fins */
	for (i = 0; i < num_clients; ++i) {
		ret = recv(connections[i], &fin, sizeof(fin), 0);
	}

	ret = 0;

out_close_conns:
	for (i = 0; i < num_clients; ++i) {
		if (connections[i] >= 0) {
			close(connections[i]);
		}
	}
out_free_all_info:
	free(all_info);
out_free_my_info:
	fca_free_rank_info(my_info);
out_close_srv:
	close(sockfd);
	return ret;
}

/**
 * Send rank information to rank0 and receive COLL_MLID
 */
int recv_comm_from_server(struct fca_context *ctx, const char *rank0_addr,
                         int rank0_port, fca_comm_desc_t *comm_desc, int node_root)
{
	struct sockaddr_in server_addr = {0,};
	void *my_info;
	int my_info_len;
	int sockfd;
	int ret = -1;
	struct hostent *ent;

	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (sockfd < 0) {
		error("socket() failed: %m");
		return sockfd;
	}

	if (node_root) {
		my_info = fca_get_rank_info(ctx, &my_info_len);
		if (!my_info) {
			error("Error getting rank info");
			goto out_close;
		}
	} else {
		my_info = NULL;
		my_info_len = 0;
	}
	
	ent = gethostbyname2(rank0_addr, AF_INET);
	server_addr.sin_family = AF_INET;
	server_addr.sin_port = htons(rank0_port);
	server_addr.sin_addr = *(struct in_addr*)ent->h_addr_list[0];
	ret = connect(sockfd, (struct sockaddr *)&server_addr, sizeof server_addr);
	if (ret < 0) {
		error("connect() failed: %m");
		goto out_free_info; 
	}

	/* Send length */
	ret = send(sockfd, &my_info_len, sizeof(my_info_len), 0);
	if (ret < 0) {
		error("send() failed: %m");
		goto out_free_info;
	}

	/* Send data */
	if (my_info_len > 0) {
		ret = send(sockfd, my_info, my_info_len, 0);
		if (ret < 0) {
			error("send() failed: %m");
			goto out_free_info;
		}
	}

	/* Receive communicator info */
	ret = recv(sockfd, comm_desc, sizeof *comm_desc, 0);
	if (ret < 0) {
		error("recv() failed: %m");
		goto out_free_info;
	} else if (ret != sizeof *comm_desc) {
		error("Got only %d bytes from rank0 (expected: %ld)",
		      ret, sizeof *comm_desc);
		ret = -EINVAL;
		goto out_free_info;
	}

	info("Got communicator, %d bytes from rank0", ret);
	ret = 0;

out_free_info:
	if (node_root) {
		fca_free_rank_info(my_info);
	}
out_close:
	close(sockfd);
	return ret;
}

/**
 * INIT2:
 * 
 * Create FCA communicator
 */
fca_comm_t *create_comm(struct options *opts, struct fca_context *ctx,
                        fca_comm_desc_t *comm_desc)
{
	fca_comm_init_spec_t spec;
	fca_comm_t *comm = NULL;
	int ret;
	
	assert(opts->rank < opts->comm_size);
	assert(opts->proc_idx < opts->num_procs);

	/* COMM_NEW protocol */
	if (opts->server_addr == NULL) {
		assert(opts->proc_idx == 0);
		ret = server_create_new_comm(ctx, opts->server_port, comm_desc, 
		                            opts->comm_size - 1);
	} else {
		ret = recv_comm_from_server(ctx, opts->server_addr, opts->server_port,
		                            comm_desc, opts->proc_idx == 0);
	}
	if (ret < 0) {
		error("Failed to create new communicator: %s", fca_strerror(ret));
		return NULL;
	}
	
	spec.desc = *comm_desc;
	spec.rank = opts->rank;
	spec.size = opts->comm_size;
	spec.proc_idx = opts->proc_idx;
	spec.num_procs = opts->num_procs;
	ret = fca_comm_init(ctx, &spec, &comm);

	if (ret < 0) {
		error("Failed to initialize communicator: %s", fca_strerror(ret));
		return NULL;
	}
	
	return comm;
}

void cleanup_comm(struct options *opts, struct fca_context *ctx,
                  fca_comm_desc_t *comm_desc, fca_comm_t *comm)
{
	fca_comm_destroy(comm);
	if (opts->server_addr == NULL) {
		fca_comm_end(ctx, comm_desc->comm_id);
	}
}


void perf_loop_start(struct timeval *tv_first,
                     struct timeval *tv_previous)
{
	timerclear(tv_first);
	timerclear(tv_previous);
}

void inline perf_loop_step(long count, long print_interval, struct timeval *tv_first,
                           struct timeval *tv_previous, long *count_previous)
{
	struct timeval tv_current, tv_diff;
	double d_num_usec, d_rtt;
	double num_usec, rtt;

	if (count % print_interval == 0) {
		gettimeofday(&tv_current, NULL);
		if (!timerisset(tv_first)) {
			*tv_first = tv_current;
		} else {
			timersub(&tv_current, tv_first, &tv_diff);
			num_usec = tv_diff.tv_usec + tv_diff.tv_sec* USEC_PER_SEC;
			rtt = num_usec / count;

			timersub(&tv_current, tv_previous, &tv_diff);
			d_num_usec = tv_diff.tv_usec + tv_diff.tv_sec * USEC_PER_SEC;
			d_rtt = d_num_usec / (count - *count_previous);

			info("[%8ld], %.2f (%.2f) rtt", count, d_rtt, rtt);
		}
		*tv_previous = tv_current;
		*count_previous = count;
	}
}

static inline int test_result(double *result, int len, double value)
{
	int i;

	for (i = 0; i < len; ++i) {
		if (result[i] != value) {
			error("Invalid result[%d], expected: %.2f, got: %.2f",
			      i, value, result[i]);
			return 0;
		}
	}
	return 1;
}

/**
 * WORK:
 * 
 * Run reduce operation in a loop and measure results
 */
void run_reduce_loop(struct options *opts, fca_comm_t *comm)
{
	struct timeval tv_first, tv_previous;
	long count_previous, count;
	fca_reduce_spec_t reduce;
	double data[opts->vec_size];
	double result[opts->vec_size];
	int ret, i;
	int dtype;


	/* Fill reduce buffer */
	for (i = 0; i < opts->vec_size; ++i) {
		data[i] = opts->data;
	}

	/* Search for corresponding fca operation */
	reduce.op = (enum fca_reduce_op_t)fca_translate_mpi_op("MPI_SUM");
	if (reduce.op < 0) {
	    error("Operation MPI_SUM unsupported");
	    return;
	}

	/* Search for corresponding fca datatype */
	dtype = fca_translate_mpi_dtype("MPI_DOUBLE");
	if (dtype < 0) {
	    error("Datatype MPI_DOUBLE unsupported");
	    return;
	} else {
		reduce.dtype = (enum fca_reduce_dtype_t)dtype;
	}

	reduce.length = opts->vec_size;
	reduce.sbuf = data;
	reduce.rbuf = result;

	info("Buffer size: %lu bytes", reduce.length * sizeof(data[0]));

	perf_loop_start(&tv_first, &tv_previous);
	count_previous = 0;
	for (count = 0; opts->message_count == 0 || count < opts->message_count;
									++count) {
		/*
		 * Run ALL_REDUCE
		 */
		ret = fca_do_all_reduce(comm, &reduce);
		if (ret < 0) {
			error("AllReduce operation failed: %s", fca_strerror(ret));
			break;
		}
		
		/*
		 * Check result
		 */
		if (opts->verify && !test_result(result, opts->vec_size, opts->result))
			break;
	
		/*
		 * Print performance
		 */
		perf_loop_step(count, opts->print_interval, &tv_first, &tv_previous,
		               &count_previous);
	}
}

void run_barrier_loop(struct options *opts, fca_comm_t *comm)
{
	struct timeval tv_first, tv_previous;
	long count_previous, count;
	int ret;
	
	perf_loop_start(&tv_first, &tv_previous);
	count_previous = 0;
	for (count = 0; opts->message_count == 0 || count < opts->message_count;
									++count) {
		/*
		 * Run BARRIER
		 */
		ret = fca_do_barrier(comm);
		if (ret < 0) {
			error("Barrier operation failed: %s", fca_strerror(ret));
			break;
		}
		
		/*
		 * Print performance
		 */
		perf_loop_step(count, opts->print_interval, &tv_first, &tv_previous,
		               &count_previous);
	}
}

void run_bcast_loop(struct options *opts, fca_comm_t *comm)
{
	struct timeval tv_first, tv_previous;
	long count_previous, count;
	double data[opts->vec_size];
	struct fca_bcast_spec bcast;
	int ret;
	int i;

	for (i = 0; i < opts->vec_size; ++i) {
		data[i] = opts->data;
	}
	
	bcast.root = 0;
	bcast.buf = data;
	bcast.size = sizeof(data);

	perf_loop_start(&tv_first, &tv_previous);
	count_previous = 0;
	for (count = 0; opts->message_count == 0 || count < opts->message_count;
									++count) {
		/*
		 * Run BCAST
		 */
		ret = fca_do_bcast(comm, &bcast);
		if (ret < 0) {
			error("Bcast operation failed: %s", fca_strerror(ret));
			break;
		}
		
		/*
		 * Check result
		 */
		if (opts->verify && !test_result(data, opts->vec_size, opts->result))
			break;

		/*
		 * Print performance
		 */
		perf_loop_step(count, opts->print_interval, &tv_first, &tv_previous,
		               &count_previous);
	}
}

/**
 * MAIN:
 * 
 */
int main(int argc, char** argv, char **envp)
{
	struct options options;
	struct fca_context *ctx;
	fca_comm_desc_t comm_desc;
	fca_comm_t *comm;
	fca_comm_caps_t comm_caps;
	
	if (parse_options(argc, argv, envp, &options)) {
		return -1;
	}
		
	ctx = init_context(options.fca_spec);
	if (!ctx)
		return -2;

	fca_free_init_spec(options.fca_spec);
	
	comm = create_comm(&options, ctx, &comm_desc);
	if (!comm) {
		fca_cleanup(ctx);
		return -3;
	}

	/* Let's see the max payload of this communicator. */
	fca_comm_get_caps(comm, &comm_caps);
	info("Comm max data payload: %d", comm_caps.max_nofrag_payload);

	switch (options.test_type) {
	case TEST_REDUCE:
		info("Running reduce...");
		run_reduce_loop(&options, comm);
		break;
	case TEST_BARRIER:
		info("Running barrier...");
		run_barrier_loop(&options, comm);
		break;
	case TEST_BCAST:
		info("Running bcast...");
		run_bcast_loop(&options, comm);
		break;
	default:
		break;
	}
	info("Test finished");

	cleanup_comm(&options, ctx, &comm_desc, comm);
	fca_cleanup(ctx);
	return 0;
}
